﻿#include "pch.h"
#include "ComputerPlayer.h"
#include <time.h>
#include "minimax.h"

using namespace TicTacToeLibrary;

using namespace concurrency;
using namespace Platform;
using namespace Windows::Foundation;
using namespace Windows::Foundation::Collections;

ComputerPlayer::ComputerPlayer(wchar_t symbol, wchar_t opponentSymbol, wchar_t emptySymbol) : m_symbol(symbol), m_opponentSymbol(opponentSymbol), m_emptySymbol(emptySymbol), m_playerType(PlayerType::Computer)
{
    m_rng.seed((unsigned)time(nullptr));
}

IAsyncOperationWithProgress<uint32, double>^ ComputerPlayer::ThinkAsync(IVector<wchar_t>^ gameBoard)
{
    return create_async([this, gameBoard](progress_reporter<double> reporter)
    {
        // Copy the contents of the IVector to a std::array.
        std::array<wchar_t, 9> board;
        std::copy(begin(gameBoard), end(gameBoard), begin(board));

        // Frequently take the center square if the board is empty.
        if (m_rng()%2 == 0 && std::count(begin(board), end(board), m_emptySymbol) == 9)
        {
            return 4u;
        }

        // Approximate counts of iterations the algorithm makes for each number of available moves at the root level.
        const std::array<uint32_t, 10> all_iterations = { 0u, 2u, 5u, 15u, 50u, 200u, 930u, 7300u, 60000u, 550000u };
        auto moves = available_moves(board, m_emptySymbol);
        const uint32_t total_iterations = all_iterations[std::count(begin(moves), end(moves), 1u)];

        // Report every 1%.
        uint32_t next_report_iter = total_iterations/100; 
        const uint32_t iter_by = next_report_iter;

        // This is called on every iteration of the minimax algorithm.
        auto callback = [total_iterations, &next_report_iter, iter_by, reporter](uint32_t iter_count)
        {
            if (iter_count == next_report_iter)
            {
                const double progress = 100.0 * iter_count / total_iterations;
                reporter.report(std::min(progress, 100.0));
                next_report_iter += iter_by;
            }
        };

        // Run the minimax algorithm in parallel if there are enough iterations.
        // Otherwise, run it serially, as the overhead of parallelism may not benefit.
        uint32_t iter_count = 0u;
        std::tuple<int32_t, int32_t> t;
        if (total_iterations > 500)
        {
            t = parallel_minimax(board, m_symbol, m_opponentSymbol, m_emptySymbol, m_symbol, iter_count, callback);
        }
        else
        {
            t = minimax(board, m_symbol, m_opponentSymbol, m_emptySymbol, m_symbol, iter_count, callback);
        }

        // Return the index part.
        return (uint32)(std::get<0>(t));
    });
}
