#include "pch.h"
#include "GameProcessor.h"
#include "DelegateCommand.h"
#include "HumanPlayer.h"
#include "ComputerPlayer.h"

using namespace TicTacToeLibrary;

using namespace concurrency;
using namespace Platform;
using namespace Platform::Collections;
using namespace std;
using namespace Windows::Foundation;
using namespace Windows::Foundation::Collections;
using namespace Windows::UI;
using namespace Windows::UI::Xaml::Media;

enum class gameResult
{
    Nought,
    Cross,
    Tie
};

// Implementation for 3x3 board.
bool checkEndOfGame(const std::array<wchar_t, 9>& gameBoard, wchar_t empty, wchar_t nought, wchar_t cross, gameResult& result, std::array<uint32_t, 3>& winningPath)
{
    /*
    |0|1|2|
    |3|4|5|
    |6|7|8|
    */
    const std::array<std::array<uint32_t, 3>, 8> permutations = 
    { {
        // Horizontals
        { {0, 1, 2} },
        { {3, 4, 5} },
        { {6, 7, 8} },

        // Verticals
        { {0, 3, 6} },
        { {1, 4, 7} },
        { {2, 5, 8} },

        // Diagonals
        { {0, 4, 8} },
        { {2, 4, 6} },
    } };

    for (auto&& permutation : permutations)
    {
        auto size = static_cast<int32_t>(permutation.size());

        if (count_if(begin(permutation), end(permutation), [gameBoard, nought](uint32_t index)
        {
            return gameBoard[index] == nought;

        }) == size)
        {
            result = gameResult::Nought;
            winningPath = permutation;
            return true;
        }

        if (count_if(begin(permutation), end(permutation), [gameBoard, cross](uint32_t index)
        {
            return gameBoard[index] == cross;

        }) == size)
        {
            result = gameResult::Cross;
            winningPath = permutation;
            return true;
        }
    }

    if (count(begin(gameBoard), end(gameBoard), empty) == 0)
    {
        result = gameResult::Tie;
        return true;
    }
    return false;
}

GameProcessor::GameProcessor()
    : m_cells(ref new Vector<Cell^>())
    , m_gameInProgress(false)
    , m_crossWins(0u)
    , m_noughtWins(0u)
    , m_backgroundProgress(0)
    , m_playerOptions(ref new Vector<String^>())
    , m_defaultButtonForegroundBrush(ref new SolidColorBrush(Colors::White))
    , m_winningPathButtonForegroundBrush(ref new SolidColorBrush(Colors::Red))
{
    for (uint32_t i = 0; i < 9; i++)
    {
        m_cells->Append(ref new Cell(i, m_defaultButtonForegroundBrush, this));
    }

    for (auto cell : m_cells)
    {
        cell->CellSelected += ref new CellSelectedHandler(this, &GameProcessor::CellSelected);
    }

    m_newGameCommand = ref new DelegateCommand(ref new ExecuteDelegate(this, &GameProcessor::CreateNewGame), ref new CanExecuteDelegate(this, &GameProcessor::CanSelectNewGame));

    m_playerOptions->Append("Human");
    m_playerOptions->Append("Computer");

    SelectedPlayerOption1 = 0;
    SelectedPlayerOption2 = 1;
}

void GameProcessor::CellSelected(Cell^ sender)
{
    ProcessTurn(sender->Index);
}

void GameProcessor::CheckEndOfGame()
{
    gameResult result;
    std::array<uint32_t, 3> path;
    if (checkEndOfGame(m_gameBoard, EmptySymbol, NoughtSymbol, CrossSymbol, result, path))
    {
        switch (result)
        {
        case gameResult::Nought:
            for (auto index : path)
            {
                m_cells->GetAt(index)->ForegroundBrush = m_winningPathButtonForegroundBrush;
            }
            m_noughtWins++;
            OnPropertyChanged("NoughtWins");
            m_gameResult = NoughtSymbol + " wins!";
            break;
        case gameResult::Cross:
            for (auto index : path)
            {
                m_cells->GetAt(index)->ForegroundBrush = m_winningPathButtonForegroundBrush;
            }
            m_crossWins++;
            OnPropertyChanged("CrossWins");
            m_gameResult = CrossSymbol + " wins!";
            break;
        case gameResult::Tie:
            m_gameResult = "It's a tie!";
            break;
        }

        m_gameInProgress = false;
        OnPropertyChanged("GameInProgress");
        OnPropertyChanged("GameResult");
        OnPropertyChanged("CanCreateNewGame");
    }
    else
    {
        SwitchPlayers();
    }
}

void GameProcessor::CreateNewGame(Object^ parameter)
{
    (void) parameter;

    const int32 Human = 0;
#ifdef _DEBUG
    const int32 Computer = 1;
#endif

    if (SelectedPlayerOption1 == Human)
    {
        m_player1 = ref new HumanPlayer(CrossSymbol);
    }
    else
    {
        assert(SelectedPlayerOption1 == Computer);
        m_player1 = ref new ComputerPlayer(CrossSymbol, NoughtSymbol, EmptySymbol);
    }

    if (SelectedPlayerOption2 == Human)
    {
        m_player2 = ref new HumanPlayer(NoughtSymbol);
    }
    else
    {
        assert(SelectedPlayerOption2 == Computer);
        m_player2 = ref new ComputerPlayer(NoughtSymbol, CrossSymbol, EmptySymbol);
    }

    // Reset cells.
    m_cells->Clear();
    for (uint32_t i = 0; i < 9; i++)
    {
        m_cells->Append(ref new Cell(i, m_defaultButtonForegroundBrush, this));
    }

    for (auto cell : m_cells)
    {
        cell->CellSelected += ref new CellSelectedHandler(this, &GameProcessor::CellSelected);
    }

    // Initialize game board state.
    fill(begin(m_gameBoard), end(m_gameBoard), EmptySymbol);

    // Hide the game setup area and previous game's result text.
    m_gameResult = "";
    OnPropertyChanged("GameResult");
    m_gameInProgress = true;
    OnPropertyChanged("GameInProgress");
    OnPropertyChanged("CanCreateNewGame");

    // Player 1 starts.
    m_currentPlayer = m_player1;

    // If player 1 is the computer, set the current to player 2 and switch them
    // to activate player 1's turn.
    if (m_currentPlayer->Player == TicTacToeLibrary::PlayerType::Computer)
    {
        m_currentPlayer = m_player2;
        SwitchPlayers();
    }
}

void GameProcessor::SwitchPlayers()
{
    // Switch player by toggling pointer.
    m_currentPlayer = (m_currentPlayer == m_player1) ? m_player2 : m_player1;

    // If the current player is computer-controlled, call the ThinkAsync method in the background, 
    // and then process the computer's move.
    if (m_currentPlayer->Player != TicTacToeLibrary::PlayerType::Computer)
    {
        return;
    }

    m_currentThinkOp = m_currentPlayer->ThinkAsync(ref new Vector<wchar_t>(m_gameBoard));
    m_currentThinkOp->Progress = ref new AsyncOperationProgressHandler<uint32, double>([this](IAsyncOperationWithProgress<uint32, double>^ asyncInfo, double value)
    {
        (void) asyncInfo; // Unused parameter

        // Update progress bar.
        m_backgroundProgress = value;
        OnPropertyChanged("BackgroundProgress");
    });

    // Create a task that wraps the async operation. After the task completes, select the 
    // cell that the computer chose.
    create_task(m_currentThinkOp).then([this](task<uint32> previousTask)
    {
        m_currentThinkOp = nullptr;

        // I use a task-based continuation here so ensure that this continuation runs to 
        // guarantee that the UI is updated. You should consider putting a try/catch block
        // around calls to task::get to handle any errors.
        uint32 move = previousTask.get();

        // Choose the cell.
        m_cells->GetAt(move)->Select(nullptr);

        // Reset the progress bar.
        m_backgroundProgress = 0.0;
        OnPropertyChanged("BackgroundProgress");

    }, task_continuation_context::use_current());
}

void GameProcessor::ProcessTurn(int32 index)
{
    m_gameBoard[index] = m_currentPlayer->Symbol;
    CheckEndOfGame();
}