#pragma once
#include <algorithm>
#include <array>
#include <ppltasks.h>
#include <stdint.h>
#include <tuple>

// Finds all available moves.
// A value of 0 means that the move is not available; 1 means that the cell is empty and a move is available.
template <size_t Size>
std::array<uint32_t, Size> available_moves(const std::array<wchar_t, Size>& board, wchar_t empty)
{
    std::array<uint32_t, Size> result;
    for (uint32_t i = 0; i < Size; i++)
    {
        result[i] = (board[i] == empty) ? 1u : 0u;
    }
    return result;
}

// Determines the winner.
// This function returns the empty parameter if the game is still active.
// Implementation for 3x3 board.
wchar_t winner(const std::array<wchar_t, 9>& board, wchar_t player, wchar_t opponent, wchar_t empty)
{
    /*
    |0|1|2|
    |3|4|5|
    |6|7|8|
    */
    const std::array<std::array<uint32_t, 3>, 8> permutations = 
    { {
        // Horizontals
        { {0, 1, 2} },
        { {3, 4, 5} },
        { {6, 7, 8} },

        // Verticals
        { {0, 3, 6} },
        { {1, 4, 7} },
        { {2, 5, 8} },

        // Diagonals
        { {0, 4, 8} },
        { {2, 4, 6} },
    } };

    for (auto&& permutation : permutations)
    {
        auto size = static_cast<int32_t>(permutation.size());

        if (std::count_if(begin(permutation), end(permutation), [board, player](uint32_t index)
        {
            return board[index] == player;

        }) == size)
        {
            return player;
        }

        if (std::count_if(begin(permutation), end(permutation), [board, opponent](uint32_t index)
        {
            return board[index] == opponent;

        }) == size)
        {
            return opponent;
        }
    }

    return empty;
}

// Determines whether the game is over.
template <size_t Size>
bool terminal(const std::array<wchar_t, Size>& board, wchar_t player, wchar_t opponent, wchar_t empty)
{
    auto moves = available_moves(board, empty);
    return std::find(begin(moves), end(moves), 1u) == end(moves) || winner(board, player, opponent, empty) != empty;
}

// Parallel version of the minimax algorithm.
template <size_t Size, typename CallbackFn>
std::tuple<int32_t, int32_t> parallel_minimax(const std::array<wchar_t, Size>& board, wchar_t player, wchar_t opponent, wchar_t empty, wchar_t max, uint32_t& iter_count, CallbackFn& iter_callback)
{
    iter_callback(++iter_count);

    auto space = std::make_tuple(-1, 0);

    auto w = winner(board, player, opponent, empty);
    if (w == empty)
    {
        std::get<1>(space) = 0;
    }
    else if (w == max)
    {
        std::get<1>(space) = +1;
    }
    else
    {
        std::get<1>(space) = -1;
    }
    if (terminal(board, player, opponent, empty))
    {
        return space;
    }

    concurrency::combinable<bool> first_moves([] { return true; });
    concurrency::combinable<std::tuple<int32_t, int32_t>> best_moves;
    auto available = available_moves(board, empty);
    auto available_size = std::count(begin(available), end(available), 1u);
    parallel_for(0u, static_cast<uint32_t>(available.size()), [&first_moves, &best_moves, &available, available_size, &board, player, opponent, empty, max, &iter_count, &iter_callback](uint32_t i)
    {
        if (available[i] == 0u)
        {
            return;
        }
        std::array<wchar_t, Size> new_board(board);
        new_board[i] = player;
        std::tuple<int32_t, int32_t> temp;
        // Choose the parallel version only if there is enough work yet to do.
        if (available_size < 7)
        {
            temp = minimax(new_board, opponent, player, empty, max, iter_count, iter_callback);
        }
        else
        {
            temp = parallel_minimax(new_board, opponent, player, empty, max, iter_count, iter_callback);
        }   
        std::get<0>(temp) = i;

        // If the new move is better than our previous move, take it.
        if (first_moves.local() || (player == max && std::get<1>(temp) > std::get<1>(best_moves.local())) || (player != max && std::get<1>(temp) < std::get<1>(best_moves.local())))
        { 
            first_moves.local() = false;
            best_moves.local() = temp; 
        }
    });

    bool first = true;
    std::tuple<int32_t, int32_t> best;
    best_moves.combine_each([&best, &first, player, max](std::tuple<int32_t, int32_t> local_result)
    {
        if (first || (player == max && std::get<1>(local_result) > std::get<1>(best)) || (player != max && std::get<1>(local_result) < std::get<1>(best)))
        { 
            first = false;
            best = local_result; 
        }
    });
    return best;
}

// Serial version of the minimax algorithm.
template <size_t Size, typename CallbackFn>
std::tuple<int32_t, int32_t> minimax(const std::array<wchar_t, Size>& board, wchar_t player, wchar_t opponent, wchar_t empty, wchar_t max, uint32_t& iter_count, CallbackFn& iter_callback)
{
    iter_callback(++iter_count);

    auto space = std::make_tuple(-1, 0);

    auto w = winner(board, player, opponent, empty);
    if (w == empty)
    {
        std::get<1>(space) = 0;
    }
    else if (w == max)
    {
        std::get<1>(space) = +1;
    }
    else
    {
        std::get<1>(space) = -1;
    }
    if (terminal(board, player, opponent, empty))
    {
        return space;
    }

    bool first = true;
    std::tuple<int32_t, int32_t> best;
    auto moves = available_moves(board, empty);
    for (uint32_t i = 0; i < moves.size(); i++)
    {
        if (moves[i])
        {
            std::array<wchar_t, Size> new_board(board);
            new_board[i] = player;
            std::tuple<int32_t, int32_t> temp = minimax(new_board, opponent, player, empty, max, iter_count, iter_callback);
            std::get<0>(temp) = i;

            // If the new move is better than our previous move, take it.
            if (first || (player == max && std::get<1>(temp) > std::get<1>(best)) || (player != max && std::get<1>(temp) < std::get<1>(best)))
            { 
                first = false;
                best = temp; 
            }
        }
    }
    return best;
}
